﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Programmable {
  [Item("ParameterVectorMoveGenerator", "Calls the GetNeighbors method of the programmable definition to obtain the moves.")]
  [StorableClass]
  public class ParameterVectorMoveGenerator : InstrumentedOperator, IParameterVectorMoveOperator, IMultiMoveGenerator, IStochasticOperator {

    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public IValueLookupParameter<IntValue> SampleSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["SampleSize"]; }
    }

    public ILookupParameter<SingleObjectiveScript> ScriptParameter {
      get { return (ILookupParameter<SingleObjectiveScript>)Parameters["Script"]; }
    }

    public ILookupParameter<Configuration> ConfigurationParameter {
      get { return (ILookupParameter<Configuration>)Parameters["Configuration"]; }
    }

    [StorableConstructor]
    protected ParameterVectorMoveGenerator(bool deserializing) : base(deserializing) { }
    protected ParameterVectorMoveGenerator(ParameterVectorMoveGenerator original, Cloner cloner)
      : base(original, cloner) { }
    public ParameterVectorMoveGenerator() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new ValueLookupParameter<IntValue>("SampleSize", "The number of moves to sample."));
      Parameters.Add(new LookupParameter<SingleObjectiveScript>("Script", "The script that will execute the evaluation function and define the parameter vector."));
      Parameters.Add(new LookupParameter<Configuration>("Configuration", "An item that holds the problem's configuration."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParameterVectorMoveGenerator(this, cloner);
    }

    public override IOperation InstrumentedApply() {
      var random = RandomParameter.ActualValue;
      var runner = ScriptParameter.ActualValue;
      if (runner.Instance == null) throw new InvalidOperationException("Script instance is null, maybe the code doesn't compile.");
      var sampleSize = SampleSizeParameter.ActualValue.Value;
      var config = ConfigurationParameter.ActualValue;
      var binDict = new Dictionary<string, BinaryVector>();
      var intDict = new Dictionary<string, IntegerVector>();
      var realDict = new Dictionary<string, RealVector>();
      var permDict = new Dictionary<string, Permutation>();
      foreach (var param in config.Parameters) {
        var binConfig = param.Value as BinaryParameterConfiguration;
        if (binConfig != null) {
          binDict.Add(param.Key, (BinaryVector)ExecutionContext.Scope.Variables[param.Key].Value);
          continue;
        }
        var intConfig = param.Value as IntegerParameterConfiguration;
        if (intConfig != null) {
          intDict.Add(param.Key, (IntegerVector)ExecutionContext.Scope.Variables[param.Key].Value);
          continue;
        }
        var realConfig = param.Value as RealParameterConfiguration;
        if (realConfig != null) {
          realDict.Add(param.Key, (RealVector)ExecutionContext.Scope.Variables[param.Key].Value);
          continue;
        }
        var permConfig = param.Value as PermutationParameterConfiguration;
        if (permConfig != null) {
          permDict.Add(param.Key, (Permutation)ExecutionContext.Scope.Variables[param.Key].Value);
          continue;
        }
        throw new InvalidOperationException("Parameter " + param.Key + " not found.");
      }
      var vector = new ParameterVector(
        binaryVectors: binDict.Count > 0 ? binDict : null,
        integerVectors: intDict.Count > 0 ? intDict : null,
        realVectors: realDict.Count > 0 ? realDict : null,
        permutations: permDict.Count > 0 ? permDict : null);
      var nbhood = runner.Instance.GetNeighbors(random, vector).Take(sampleSize).ToList();

      var moveScopes = new Scope[nbhood.Count];
      for (int i = 0; i < moveScopes.Length; i++) {
        moveScopes[i] = new Scope(i.ToString());
        foreach (var param in nbhood[i].BinaryNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].Binary(param)));

        foreach (var param in nbhood[i].IntegerNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].Integer(param)));

        foreach (var param in nbhood[i].RealNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].Real(param)));

        foreach (var param in nbhood[i].PermutationNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].Permutation(param)));
      }
      ExecutionContext.Scope.SubScopes.AddRange(moveScopes);

      return base.InstrumentedApply();
    }
  }
}
